/*
 *  Copyright 2007 CNRS-AIST
 *
 *  Authors: Fumio Kanehiro, Francois Keith, Florent Lamiraux
 */

#ifndef HRP2_DYNAMICS_HRP2OPTHUMANOIDDYNAMICROBOT_HH
# define HRP2_DYNAMICS_HRP2OPTHUMANOIDDYNAMICROBOT_HH
# include <jrl/mal/matrixabstractlayer.hh>
# include <jrl/dynamics/humanoiddynamicrobot.hh>
# include <jrl/dynamics/dynamicsfactory.hh>

# include <hrp2-dynamics/config.hh>

/// \brief Optimized implementation of HRP2 dynamic model
class HRP2_DYNAMICS_DLLAPI Chrp2OptHumanoidDynamicRobot
: public virtual jrlDelegate::humanoidDynamicRobot
{
 public:
  /// \brief Constructor.
  Chrp2OptHumanoidDynamicRobot
    (CjrlRobotDynamicsObjectFactory *anObjectFactory);

  /// \brief Default constructor.
  Chrp2OptHumanoidDynamicRobot();

  /// \brief Destructor.
  virtual ~Chrp2OptHumanoidDynamicRobot();

  /// \brief Call parent implementation and initialize exact inverse
  ///        kinematics.
  virtual bool initialize();

  /** Inherited- see parent class for doc */
  virtual double upperBoundDof(unsigned int inRankInConfiguration,
			       const vectorN& inConfig);

  /** Inherited- see parent class for doc */
  virtual double lowerBoundDof(unsigned int inRankInConfiguration,
			       const vectorN& inConfig);

  /** Inherited- see parent class for doc */
  virtual bool setHandClench(CjrlHand* inHand, double inClenchingValue);

  /** Inherited- see parent class for doc */
  virtual double getHandClench(CjrlHand* inHand);

  /** Inherited- see parent class for doc */
  virtual void computeJacobianCenterOfMass();

  /** Inherited- see parent class for doc */
  virtual void computeInertiaMatrix();

  /** Inherited- see parent class for doc */
  virtual bool getJacobianCenterOfMass(const CjrlJoint& inStartJoint,
				       matrixNxP& outjacobian,
				       unsigned int offset = 0,
				       bool inIncludeStartFreeFlyer = true);

  /**
     \brief set the order of joints. default value is false.
     \param i_flag true means joints are ordered as follows.
     root, right_leg, left_leg, waist, neck, right_arm, right_hand,
     left_arm, left_hand. false does as follows. root, right_leg,
     left_leg, waist, neck, right_arm, left_arm, right_hand, left_hand.
  */
  void isKineoOrder(bool i_flag);

  /// \brief \brief Returns the inertia Matrix of HRP-2.
  const matrixNxP & inertiaMatrix() const;

  /// \brief Returns the Jacobian of the centern of Mass.
  const MAL_MATRIX_(double) & jacobianCenterOfMass() const;

  /** \brief Computes analytical inverse cinematics for some sub-parts of HRP-2.

      The sub-parts considered in this implementation are: the legs
      and the arms.
  */
  bool getSpecializedInverseKinematics(const CjrlJoint& jointRoot,
				       const CjrlJoint& jointEnd,
				       const matrix4d& jointRootPosition,
				       const matrix4d& jointEndPosition,
				       vectorN &q);

 protected:

  /*! \brief Compute the analytical Inverse Kinematics between the
    waist and the legs. */
  void getWaistFootKinematics(const matrix4d& jointRootPosition,
			      const matrix4d& jointEndPosition,
			      vectorN &q,
			      vector3d Dt);

  /*! \brief Compute the analytical Inverse Kinematics between the
    waist and the legs. */
  void getShoulderWristKinematics(const matrix4d& jointRootPosition,
				  const matrix4d& jointEndPosition,
				  vectorN& q,
				  int side);

  /*! \brief Detect automatically shoulders.
    This is done by calling appropriatly DetectAutomaticallyOneShoulder.
  */
  void DetectAutomaticallyShoulders();


  /*! \brief Detect automatically one shoulder.
    Here we simply assume that the shoulder is after the chest
    along the kinematic chain from the root to the hand.
    \param[in] aHand: The hand which belongs to the arm for which we wants
    to find the shoulder.
    \param[out] aShoulder: Returns the pointer to the joint which
    handles the joint.
  */
  void DetectAutomaticallyOneShoulder(CjrlHand* aHand,
				      CjrlJoint*& aShoulder);

  /// \brief Compute the displacement X max for a given height
  double ComputeXmax(double& Z);

 private:
  // This should never be copied.
  // To avoid that, these method are NOT implemented.
  // If you have undefined symbol(s), you are doing something wrong...
  Chrp2OptHumanoidDynamicRobot (const Chrp2OptHumanoidDynamicRobot&);
  Chrp2OptHumanoidDynamicRobot&
    operator=(const Chrp2OptHumanoidDynamicRobot&);


  /// \brief Define DoF order.
  bool m_isKineoOrder;

  /// Jacobian of the center of mass.
  matrixNxP m_jcom;

  /// Inertia matrix.
  matrixNxP m_InertiaMatrix;

  /// \brief Left should joint.
  CjrlJoint* m_LeftShoulder;

  /// \brief Right should joint.
  CjrlJoint* m_RightShoulder;
};

// ranks in configuration
# define RLEG_JOINT1  (6+ 1)
# define RLEG_JOINT2  (6+ 2)
# define RLEG_JOINT4  (6+ 4)
# define RLEG_JOINT5  (6+ 5)
# define LLEG_JOINT1  (6+ 7)
# define LLEG_JOINT2  (6+ 8)
# define LLEG_JOINT4  (6+10)
# define LLEG_JOINT5  (6+11)
# define CHEST_JOINT0 (6+12)
# define CHEST_JOINT1 (6+13)
# define LEFT_HAND    (6+16+12+7)
# define RIGHT_ARM    (6+16)

#endif //! HRP2_DYNAMICS_HRP2OPTHUMANOIDDYNAMICROBOT_HH
