/*
 *  Copyright CNRS-AIST
 *
 *  Authors: Fumio Kanehiro, Florent Lamiraux, Olivier Stasse
 */

#include <iostream>
#include "hrp2-dynamics/hrp2OptHumanoidDynamicRobot.h"

using namespace jrlDelegate;

// ranks in configuration
#define RLEG_JOINT1  (6+ 1)
#define RLEG_JOINT2  (6+ 2)
#define RLEG_JOINT4  (6+ 4)
#define RLEG_JOINT5  (6+ 5)
#define LLEG_JOINT1  (6+ 7)
#define LLEG_JOINT2  (6+ 8)
#define LLEG_JOINT4  (6+10)
#define LLEG_JOINT5  (6+11)
#define CHEST_JOINT0 (6+12)
#define CHEST_JOINT1 (6+13)
#define LEFT_HAND    (6+16+12+7)
#define RIGHT_ARM    (6+16)

Chrp2OptHumanoidDynamicRobot::~Chrp2OptHumanoidDynamicRobot()
{
}

Chrp2OptHumanoidDynamicRobot::Chrp2OptHumanoidDynamicRobot
(CjrlRobotDynamicsObjectFactory *anObjectFactory)
  : humanoidDynamicRobot(anObjectFactory),
    m_isKineoOrder (false),
    m_jcom (),
    m_InertiaMatrix (),
    m_LeftShoulder (),
    m_RightShoulder ()
{}

bool Chrp2OptHumanoidDynamicRobot::initialize()
{
  if (jrlDelegate::humanoidDynamicRobot::initialize() == false) {
    return false;
  }
  DetectAutomaticallyShoulders();
  return true;
}

void Chrp2OptHumanoidDynamicRobot::isKineoOrder(bool i_flag)
{
    m_isKineoOrder = i_flag;
}

double Chrp2OptHumanoidDynamicRobot::upperBoundDof(unsigned int inRankInConfiguration,
						   const vectorN& inConfig)
{
    double q, limit, ub;
    ub = humanoidDynamicRobot::upperBoundDof(inRankInConfiguration);
    if (inRankInConfiguration==CHEST_JOINT0){
	q = inConfig(CHEST_JOINT1);
	if (q > 0.76){
	    limit = tan(-(q-0.91)/0.1)/18+0.2;
	    if (limit < ub) ub = limit;
	}
    }else if (inRankInConfiguration==LLEG_JOINT1){
	q = inConfig(LLEG_JOINT2);
	limit = 0.2182*atan((q+1.6406)*5.7296)+0.27925;
	if (limit < ub) ub = limit;
    }else if (inRankInConfiguration==CHEST_JOINT1){
	q = inConfig(CHEST_JOINT0);
	if(q<0)
	    limit =  0.1*atan((q+0.2)*18)+0.91;
	else
	    limit = -0.1*atan((q-0.2)*18)+0.91;
	if (limit < ub) ub = limit;
    }else if (inRankInConfiguration==LLEG_JOINT5){
	q = inConfig(LLEG_JOINT4);
	limit = -0.0024/((q+1.4)*(q+1.4))+0.35;
	if (limit < ub) ub = limit;
    }
    return ub;
}

double Chrp2OptHumanoidDynamicRobot::lowerBoundDof(unsigned int inRankInConfiguration,
						   const vectorN& inConfig)
{
    double q, limit, lb;
    lb = humanoidDynamicRobot::lowerBoundDof(inRankInConfiguration);
    if (inRankInConfiguration==RLEG_JOINT1){
	q = inConfig(RLEG_JOINT2);
	limit = -0.2182*atan((q+1.6406)*5.7296)-0.27925;
	if (limit > lb) lb = limit;
    }else if (inRankInConfiguration==RLEG_JOINT2){
	q = inConfig(RLEG_JOINT1);
	if (q < 0.05){
	    limit = tan(-(q+0.27925)/0.2182)/5.7296-1.6406;
	    if (limit > lb) lb = limit;
	}
    }else if (inRankInConfiguration==LLEG_JOINT2){
	q = inConfig(LLEG_JOINT1);
	if (q > 0.05){
	    limit = tan((q-0.27925)/0.2182)/5.7296-1.6406;
	    if (limit > lb) lb = limit;
	}
    }else if (inRankInConfiguration==LLEG_JOINT4){
	q = inConfig(LLEG_JOINT5);
	limit = sqrt( 0.0024/(q+0.35))-1.4;
	if (limit > lb) lb = limit;
    }else if (inRankInConfiguration==RLEG_JOINT5){
	q = inConfig(RLEG_JOINT4);
	limit =  0.0024/((q+1.4)*(q+1.4))-0.35;
	if (limit > lb) lb = limit;
    }else if (inRankInConfiguration==LLEG_JOINT4){
	q = inConfig(LLEG_JOINT5);
	limit = sqrt(-0.0024/(q-0.35))-1.4;
	if (limit > lb) lb = limit;
    }else if (inRankInConfiguration==CHEST_JOINT0){
	q = inConfig(CHEST_JOINT1);
	if (q > 0.76){
	    limit = tan( (q-0.91)/0.1)/18-0.2;
	    if (limit > lb) lb = limit;
	}
    }
    return lb;
}

double Chrp2OptHumanoidDynamicRobot::getHandClench(CjrlHand* inHand)
{

    //Designed for the parallel mechanism of HRP2 two-jaw hands
    unsigned int dof = inHand->associatedWrist()->childJoint(0)->rankInConfiguration();
    double upperLimit = humanoidDynamicRobot::upperBoundDof(dof);
    double lowerLimit = humanoidDynamicRobot::lowerBoundDof(dof);
    double curVal = currentConfiguration()(dof);

    return (upperLimit - curVal)/(upperLimit-lowerLimit);
}

bool Chrp2OptHumanoidDynamicRobot::setHandClench(CjrlHand* inHand,
						 double inClenchingValue)
{
    //Designed for the parallel mechanism of HRP2 two-jaw hands
    if (inClenchingValue > 1 || inClenchingValue < 0)
    {
        std::cout << "Chrp2OptHumanoidDynamicRobot::setHandClench(): argument 2 should be between 0 and 1\n";
        return false;
    }

    unsigned int dof = inHand->associatedWrist()->childJoint(0)->rankInConfiguration();
    double upperLimit = humanoidDynamicRobot::upperBoundDof(dof);
    double lowerLimit = humanoidDynamicRobot::lowerBoundDof(dof);
    double wantedVal =  upperLimit - inClenchingValue*(upperLimit-lowerLimit);

    vectorN config = currentConfiguration();

    config(dof) = wantedVal;

    unsigned int parallelDofStart;
    if (inHand == rightHand())
	if (m_isKineoOrder){
	    parallelDofStart = 6+16+7;
	}else{
	    parallelDofStart = 6+16+7+7;
	}
    else
        parallelDofStart = LEFT_HAND;

    int coef = -1;
    for (unsigned int i = parallelDofStart ;i<parallelDofStart+5; i++)
    {
        config(i) = coef * wantedVal;
        coef *= -1;
    }

    currentConfiguration( config );
    return true;
}



const matrixNxP & Chrp2OptHumanoidDynamicRobot::inertiaMatrix() const
{
  return m_InertiaMatrix;
}

const MAL_MATRIX_(double) & Chrp2OptHumanoidDynamicRobot::jacobianCenterOfMass() const
{
  return m_jcom;

}
